"""
    File: strategic_axioms.py
    Author: Wesley H. Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: March 16, 2025
    
    Strategic axioms 
"""

import numpy as np
import math
from pref_voting.axiom import Axiom
from pref_voting.axiom_helpers import *
from itertools import permutations, combinations
from pref_voting.helper import weak_orders
from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.rankings import Ranking

def has_strategy_proofness_violation(prof, vm, set_preference = "single-winner", verbose=False):
    """
    Returns True if there is a voter who can benefit by misrepresenting their preferences.

    If set_preference = "single-winner", a voter benefits only if they can change the unique winner in the original profile to a unique winner in the new profile such that in their original ranking, the new winner is above the old winner.

    If set_preference = "weak-dominance", a voter benefits only if in their original ranking, all new winners are weakly above all old winners and some new winner is strictly above some old winner.

    If set_preference = "optimist", a voter benefits only if in their original ranking, their favorite new winner is above their favorite old winner.

    If set_preference = "pessimist", a voter benefits only if in their original ranking, their least favorite new winner is above their least favorite old winner.

    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        Result of the test (bool): Returns True if there is a violation and False otherwise.

    .. note::

        The different set preference notions are drawn from Definition 2.1.1 (p. 42) of The Mathematics of Manipulation by Alan D. Taylor. 
    """

    winners = vm(prof)

    if isinstance(prof,ProfileWithTies):
        prof.use_extended_strict_preference()

    found_manipulator = False

    ranking_tokens = prof.rankings
    ranking_types = prof.ranking_types

    ws = vm(prof)

    if set_preference == "single-winner":
        if len(ws) > 1:
            return False


    for r in ranking_types:
        if not found_manipulator:

            ranking_tokens_minus_r = [r for r in ranking_tokens]
            ranking_tokens_minus_r.remove(r)

            if isinstance(prof,Profile):

                for new_r in permutations(prof.candidates):
                    if new_r != r and not found_manipulator:

                        new_ranking_tokens = ranking_tokens_minus_r + [new_r]
                        new_prof = Profile(new_ranking_tokens)
                        new_ws = vm(new_prof)

                        old_winner_to_compare = None
                        new_winner_to_compare = None

                        if set_preference == "single-winner" and len(new_ws) == 1:

                            old_winner_to_compare = ws[0]
                            new_winner_to_compare = new_ws[0] 
                        
                        elif set_preference == "weak-dominance":
                            r_as_ranking = Ranking({c: i for i, c in enumerate(r)})
                        
                        elif set_preference == "optimist":
                                
                            old_winner_to_compare = [cand for cand in r if cand in ws][0]
                            new_winner_to_compare = [cand for cand in r if cand in new_ws][0]

                        elif set_preference == "pessimist":
                            
                            old_winner_to_compare = [cand for cand in r if cand in ws][-1] 
                            new_winner_to_compare = [cand for cand in r if cand in new_ws][-1]

                        if old_winner_to_compare is not None and r.index(old_winner_to_compare) > r.index(new_winner_to_compare) or (set_preference == "weak-dominance" and r_as_ranking.weak_dom(new_ws,ws)):
                            
                            found_manipulator = True

                            if verbose:
                                print(f"Violation of Strategy-Proofness for {vm.name} under the {set_preference} set preference.")
                                print(f"A voter can benefit by changing their ranking from {r} to {new_r}.")
                                print("")
                                print("Original Profile:")
                                prof.display()
                                print(prof.description())
                                print("")
                                vm.display(prof)
                                prof.display_margin_graph()
                                print("")
                                print("New Profile:")
                                new_prof.display()
                                print(new_prof.description())
                                print("")
                                vm.display(new_prof)
                                new_prof.display_margin_graph()

            if isinstance(prof,ProfileWithTies):
                r_dict = r.rmap

                for _new_r in weak_orders(prof.candidates):
                    new_r = Ranking(_new_r)
                    if new_r != r and not found_manipulator:

                        new_ranking_tokens = ranking_tokens_minus_r + [new_r]
                        new_prof = ProfileWithTies(new_ranking_tokens, candidates = prof.candidates)
                        new_prof.use_extended_strict_preference()
                        new_ws = vm(new_prof)

                        ranked_old_winners = [c for c in ws if c in r_dict.keys()]
                        ranked_new_winners = [c for c in new_ws if c in r_dict.keys()]

                        rank_of_old_winner_to_compare = None
                        rank_of_new_winner_to_compare = None

                        if set_preference == "single-winner" and len(new_ws) == 1:

                            rank_of_old_winner_to_compare = r_dict[ws[0]] if ranked_old_winners else math.inf
                            rank_of_new_winner_to_compare = r_dict[new_ws[0]] if ranked_new_winners else math.inf
                        
                        elif set_preference == "optimist":

                            rank_of_old_winner_to_compare = min([r_dict[c] for c in ranked_old_winners]) if ranked_old_winners else math.inf
                            rank_of_new_winner_to_compare = min([r_dict[c] for c in ranked_new_winners]) if ranked_new_winners else math.inf

                        elif set_preference == "pessimist":

                            rank_of_old_winner_to_compare = max([r_dict[c] for c in ranked_old_winners]) if ranked_old_winners == ws else math.inf
                            rank_of_new_winner_to_compare = max([r_dict[c] for c in ranked_new_winners]) if ranked_new_winners == new_ws else math.inf

                        if rank_of_old_winner_to_compare is not None and rank_of_old_winner_to_compare > rank_of_new_winner_to_compare or (set_preference == "weak-dominance" and r.weak_dom(new_ws,ws,use_extended_preferences=True)):
                            
                            found_manipulator = True

                            if verbose:
                                print(f"Violation of Strategy-Proofness for {vm.name} under the {set_preference} set preference.")
                                print(f"A voter can benefit by changing their ranking from {r} to {new_r}.")
                                print("")
                                print("Original Profile:")
                                prof.display()
                                print(prof.description())
                                print("")
                                vm.display(prof)
                                prof.display_margin_graph()
                                print("")
                                print("New Profile:")
                                new_prof.display()
                                print(new_prof.description())
                                print("")
                                vm.display(new_prof)
                                new_prof.display_margin_graph()
        
    return found_manipulator

def find_all_strategy_proofness_violations(prof, vm, set_preference = "single-winner", verbose=False):
    """
    Returns a list of tuples (old_ranking, new_ranking) where old_ranking is the original ranking and new_ranking is the ranking that the voter can change to in order to benefit.

    If set_preference = "single-winner", a voter benefits only if they can change the unique winner in the original profile to a unique winner in the new profile such that in their original ranking, the new winner is above the old winner.

    If set_preference = "weak-dominance", a voter benefits only if in their original ranking, all new winners are weakly above all old winners and some new winner is strictly above some old winner.

    If set_preference = "optimist", a voter benefits only if in their original ranking, their favorite new winner is above their favorite old winner.

    If set_preference = "pessimist", a voter benefits only if in their original ranking, their least favorite new winner is above their least favorite old winner.

    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.

    Returns:
        A List of tuples (old_ranking, new_ranking) where old_ranking is the original ranking and new_ranking is the ranking that the voter can change to in order to benefit.
    """

    winners = vm(prof)

    if isinstance(prof,ProfileWithTies):
        prof.use_extended_strict_preference()

    violations = list()

    ranking_tokens = prof.rankings
    ranking_types = prof.ranking_types

    ws = vm(prof)

    if set_preference == "single-winner":
        if len(ws) > 1:
            return violations

    for r in ranking_types:
        ranking_tokens_minus_r = [r for r in ranking_tokens]
        ranking_tokens_minus_r.remove(r)

        if isinstance(prof,Profile):

            for new_r in permutations(prof.candidates):
                if new_r != r:

                    new_ranking_tokens = ranking_tokens_minus_r + [new_r]
                    new_prof = Profile(new_ranking_tokens)
                    new_ws = vm(new_prof)

                    old_winner_to_compare = None
                    new_winner_to_compare = None

                    if set_preference == "single-winner" and len(new_ws) == 1:

                        old_winner_to_compare = ws[0]
                        new_winner_to_compare = new_ws[0] 
                    
                    elif set_preference == "weak-dominance":
                        r_as_ranking = Ranking({c: i for i, c in enumerate(r)})
                    
                    elif set_preference == "optimist":
                            
                        old_winner_to_compare = [cand for cand in r if cand in ws][0]
                        new_winner_to_compare = [cand for cand in r if cand in new_ws][0]

                    elif set_preference == "pessimist":
                        
                        old_winner_to_compare = [cand for cand in r if cand in ws][-1] 
                        new_winner_to_compare = [cand for cand in r if cand in new_ws][-1]

                    if old_winner_to_compare is not None and r.index(old_winner_to_compare) > r.index(new_winner_to_compare) or (set_preference == "weak-dominance" and r_as_ranking.weak_dom(new_ws,ws)):
                        
                        violations.append((r,new_r))

                        if verbose:
                            print(f"Violation of Strategy-Proofness for {vm.name} under the {set_preference} set preference.")
                            print(f"A voter can benefit by changing their ranking from {r} to {new_r}.")
                            print("")
                            print("Original Profile:")
                            prof.display()
                            print(prof.description())
                            print("")
                            vm.display(prof)
                            prof.display_margin_graph()
                            print("")
                            print("New Profile:")
                            new_prof.display()
                            print(new_prof.description())
                            print("")
                            vm.display(new_prof)
                            new_prof.display_margin_graph()

        if isinstance(prof,ProfileWithTies):
            r_dict = r.rmap

            for _new_r in weak_orders(prof.candidates):
                new_r = Ranking(_new_r)
                if new_r != r:

                    new_ranking_tokens = ranking_tokens_minus_r + [new_r]
                    new_prof = ProfileWithTies(new_ranking_tokens, candidates = prof.candidates)
                    new_prof.use_extended_strict_preference()
                    new_ws = vm(new_prof)

                    ranked_old_winners = [c for c in ws if c in r_dict.keys()]
                    ranked_new_winners = [c for c in new_ws if c in r_dict.keys()]

                    rank_of_old_winner_to_compare = None
                    rank_of_new_winner_to_compare = None

                    if set_preference == "single-winner" and len(new_ws) == 1:

                        rank_of_old_winner_to_compare = r_dict[ws[0]] if ranked_old_winners else math.inf
                        rank_of_new_winner_to_compare = r_dict[new_ws[0]] if ranked_new_winners else math.inf

                    elif set_preference == "optimist":

                        rank_of_old_winner_to_compare = min([r_dict[c] for c in ranked_old_winners]) if ranked_old_winners else math.inf
                        rank_of_new_winner_to_compare = min([r_dict[c] for c in ranked_new_winners]) if ranked_new_winners else math.inf

                    elif set_preference == "pessimist":

                        rank_of_old_winner_to_compare = max([r_dict[c] for c in ranked_old_winners]) if ranked_old_winners == ws else math.inf
                        rank_of_new_winner_to_compare = max([r_dict[c] for c in ranked_new_winners]) if ranked_new_winners == new_ws else math.inf

                    if rank_of_old_winner_to_compare is not None and rank_of_old_winner_to_compare > rank_of_new_winner_to_compare or (set_preference == "weak-dominance" and r.weak_dom(new_ws,ws,use_extended_preferences=True)):

                        violations.append((r.rmap,new_r.rmap))

                        if verbose:
                            print(f"Violation of Strategy-Proofness for {vm.name} under the {set_preference} set preference.")
                            print(f"A voter can benefit by changing their ranking from {r} to {new_r}.")
                            print("")
                            print("Original Profile:")
                            prof.display()
                            print(prof.description())
                            print("")
                            vm.display(prof)
                            prof.display_margin_graph()
                            print("")
                            print("New Profile:")
                            new_prof.display()
                            print(new_prof.description())
                            print("")
                            vm.display(new_prof)
                            new_prof.display_margin_graph()

    return violations

strategy_proofness = Axiom(
    "Strategy Proofness",
    has_violation = has_strategy_proofness_violation,
    find_all_violations = find_all_strategy_proofness_violations, 
)

def truncate_ballot_from_bottom(ranking, num_to_keep):
    """
    Truncate a ballot by keeping only the top num_to_keep candidates.
    
    Args:
        ranking: A Ranking object or tuple representing a ballot
        num_to_keep: Number of top candidates to keep
        
    Returns:
        A truncated Ranking object
    """
    if isinstance(ranking, tuple):
        # For tuple rankings, keep only the first num_to_keep candidates
        truncated = ranking[:num_to_keep]
        return Ranking({c: i+1 for i, c in enumerate(truncated)})
    else:
        # For Ranking objects, sort candidates by rank and keep only the top num_to_keep
        sorted_candidates = sorted(ranking.rmap.keys(), key=lambda c: ranking.rmap[c])
        truncated_rmap = {c: ranking.rmap[c] for c in sorted_candidates[:num_to_keep]}
        return Ranking(truncated_rmap)

def has_later_no_harm_violation(prof, vm, verbose=False, coalition_size=1, uniform_coalition=True, require_resoluteness=False):
    """
    Returns True if there is a ballot (or collection of ballots) such that by truncating it from the bottom, 
    a candidate who is ranked by the truncated ballot goes from losing to winning.

    Viewed in reverse, this means that adding previously unranked candidates to the bottom of a ballot harmed a higher ranked candidate.
    
    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.
        coalition_size (int, default=1): Size of the coalition of voters who truncate their ballots.
        uniform_coalition (bool, default=True): If True, all voters in the coalition have the same ballot.
        require_resoluteness (bool, default=False): If True, only profiles with a unique winner before and after truncation are considered.
        
    Returns:
        Result of the test (bool): Returns True if there is a violation and False otherwise.
    """
    # Get the current winners
    winners = vm(prof)
    
    # Convert numpy array winners to list if needed
    if isinstance(winners, np.ndarray):
        winners = winners.tolist()
    
    # If require_resoluteness is True, skip profiles with multiple winners before truncation
    if require_resoluteness and len(winners) > 1:
        return False
    
    # For individual voter case or uniform coalition
    if uniform_coalition:
        # Check each ranking type in the profile
        for r in prof.ranking_types:
            # Skip if there aren't enough voters with this ranking for the coalition
            if isinstance(r, tuple):
                if prof.rankings.count(r) < coalition_size:
                    continue
                ranked_candidates = list(r)
                r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
            else:
                if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < coalition_size:
                    continue
                ranked_candidates = list(r.cands)
                r_as_ranking = r
            
            # Try truncating at different positions (keep only first i candidates)
            for i in range(1, len(ranked_candidates)):
                # Create truncated ballot keeping only the first i candidates
                # This is proper truncation - keeping the top i candidates in their original order
                truncated_r = truncate_ballot_from_bottom(r_as_ranking, i)
                
                # Skip if truncation didn't change anything or if truncated ballot is empty
                # Ensure at least one candidate remains ranked
                if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                    continue
                
                # Get the truncated ballot as a ranking map
                truncated_rmap = truncated_r.rmap
                
                # Create a new profile with the truncated ballot(s)
                modified_rankings = []
                
                # Add all ballots except those we'll truncate
                for ballot in prof.rankings:
                    # Skip the ballots we'll truncate
                    if isinstance(ballot, tuple) and ballot == r:
                        continue
                    elif isinstance(ballot, Ranking) and isinstance(r, Ranking) and ballot.cands == r.cands:
                        continue
                    
                    # Add the ballot in the correct format
                    if isinstance(ballot, tuple):
                        modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                    else:
                        modified_rankings.append(ballot.rmap)
                
                # Make sure we've skipped the right number of ballots
                if isinstance(r, tuple):
                    original_count = prof.rankings.count(r)
                else:
                    original_count = sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands)
                
                # Add back any ballots we shouldn't have truncated
                for _ in range(original_count - coalition_size):
                    modified_rankings.append(r_as_ranking.rmap)
                
                # Add the truncated ballots
                for _ in range(coalition_size):
                    modified_rankings.append(truncated_rmap)
                
                # Create the new profile with ties
                new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                    new_prof.use_extended_strict_preference()
                
                # Get the new winners
                new_winners = vm(new_prof)
                
                # Convert numpy array winners to list if needed
                if isinstance(new_winners, np.ndarray):
                    new_winners = new_winners.tolist()
                
                # If require_resoluteness is True, skip profiles with multiple winners after truncation
                if require_resoluteness and len(new_winners) > 1:
                    continue
                
                # Check for Later No Harm violation: a candidate ranked in the truncated ballot
                # goes from losing to winning
                truncated_candidates = truncated_r.cands
                
                # Find candidates that are ranked in the truncated ballot and went from losing to winning
                new_winners_in_truncated = [c for c in new_winners if c in truncated_candidates and c not in winners]
                
                if new_winners_in_truncated:
                    if verbose:
                        print(f"Later No Harm violation: Candidate(s) {new_winners_in_truncated} went from losing to winning due to bottom truncation.")
                        print("")
                        print(f"Original winners: {winners}")
                        print(f"New winners: {new_winners}")
                        print(f"Original ballot: {r}")
                        print(f"Truncated ballot: {truncated_r}")
                        print("")
                        print("Original profile:")
                        prof.display()
                        prof.display_margin_graph()
                        vm.display(prof)
                        print("")
                        print("Modified profile:")
                        new_prof.display()
                        new_prof.display_margin_graph()
                        vm.display(new_prof)
                    return True
    
    # For non-uniform coalition
    if not uniform_coalition and coalition_size > 1:
        # Get all possible combinations of ranking types
        ranking_combinations = list(combinations(prof.ranking_types, coalition_size))
        
        for ranking_combo in ranking_combinations:
            # Check if we have enough of each ranking type
            valid_combo = True
            for r in ranking_combo:
                if isinstance(r, tuple):
                    if prof.rankings.count(r) < ranking_combo.count(r):
                        valid_combo = False
                        break
                else:
                    if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < ranking_combo.count(r):
                        valid_combo = False
                        break
            
            if not valid_combo:
                continue
            
            # For each ranking in the combination, try truncating it
            for i, r in enumerate(ranking_combo):
                if isinstance(r, tuple):
                    ranked_candidates = list(r)
                    r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
                else:
                    ranked_candidates = list(r.cands)
                    r_as_ranking = r
                
                # Try truncating at different positions
                for j in range(1, len(ranked_candidates)):
                    # Create truncated ballot keeping only the first j candidates
                    # This is proper truncation - keeping the top j candidates in their original order
                    truncated_r = truncate_ballot_from_bottom(r_as_ranking, j)
                    
                    # Skip if truncation didn't change anything or if truncated ballot is empty
                    if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                        continue
                    
                    # Get the truncated ballot as a ranking map
                    truncated_rmap = truncated_r.rmap
                    
                    # Create a new profile with the truncated ballot(s)
                    modified_rankings = []
                    
                    # Add all ballots except those in the coalition
                    for ballot in prof.rankings:
                        skip = False
                        for coalition_r in ranking_combo:
                            if (isinstance(ballot, tuple) and ballot == coalition_r) or (isinstance(ballot, Ranking) and isinstance(coalition_r, Ranking) and ballot.cands == coalition_r.cands):
                                skip = True
                                break
                        
                        if skip:
                            continue
                        
                        # Add the ballot in the correct format
                        if isinstance(ballot, tuple):
                            modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                        else:
                            modified_rankings.append(ballot.rmap)
                    
                    # Add back the coalition ballots, with the i-th one truncated
                    for k, coalition_r in enumerate(ranking_combo):
                        if k == i:
                            modified_rankings.append(truncated_rmap)
                        else:
                            if isinstance(coalition_r, tuple):
                                modified_rankings.append({c: i+1 for i, c in enumerate(coalition_r)})
                            else:
                                modified_rankings.append(coalition_r.rmap)
                    
                    # Create the new profile with ties
                    new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                    if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                        new_prof.use_extended_strict_preference()
                    
                    # Get the new winners
                    new_winners = vm(new_prof)
                    
                    # Convert numpy array winners to list if needed
                    if isinstance(new_winners, np.ndarray):
                        new_winners = new_winners.tolist()
                    
                    # If require_resoluteness is True, skip profiles with multiple winners after truncation
                    if require_resoluteness and len(new_winners) > 1:
                        continue
                    
                    # Check for Later No Harm violation: a candidate ranked in the truncated ballot
                    # goes from losing to winning
                    truncated_candidates = truncated_r.cands
                    
                    # Find candidates that are ranked in the truncated ballot and went from losing to winning
                    new_winners_in_truncated = [c for c in new_winners if c in truncated_candidates and c not in winners]
                    
                    if new_winners_in_truncated:
                        if verbose:
                            print(f"Later No Harm violation: Candidate(s) {new_winners_in_truncated} went from losing to winning due to bottom truncation.")
                            print("")
                            print(f"Original winners: {winners}")
                            print(f"New winners: {new_winners}")
                            print(f"Original ballot: {r}")
                            print(f"Truncated ballot: {truncated_r}")
                            print("")
                            print("Original profile:")
                            prof.display()
                            prof.display_margin_graph()
                            vm.display(prof)
                            print("")
                            print("Modified profile:")
                            new_prof.display()
                            new_prof.display_margin_graph()
                            vm.display(new_prof)
                        return True
    
    return False

def find_all_later_no_harm_violations(prof, vm, verbose=False, coalition_size=1, uniform_coalition=True, require_resoluteness=False):
    """
    Returns a list of tuples (original_ballot, truncated_ballot, original_winners, new_winners, new_winners_in_truncated) such that bottom-truncating the original_ballot to the truncated_ballot causes a candidate ranked in the truncated ballot to go from losing to winning.
    
    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.
        coalition_size (int, default=1): Size of the coalition of voters who truncate their ballots.
        uniform_coalition (bool, default=True): If True, all voters in the coalition have the same ballot.
        require_resoluteness (bool, default=False): If True, only profiles with a unique winner before and after truncation are considered.
        
    Returns:
        A list of tuples (original_ballot, truncated_ballot, original_winners, new_winners, new_winners_in_truncated) 
        witnessing violations of Later No Harm. The new_winners_in_truncated element contains the candidates that 
        specifically caused the violation by going from losing to winning while being ranked in the truncated ballot.
    """
    violations = []
    
    # Get the current winners
    winners = vm(prof)
    
    # Convert numpy array winners to list if needed
    if isinstance(winners, np.ndarray):
        winners = winners.tolist()
    
    # If require_resoluteness is True, skip profiles with multiple winners before truncation
    if require_resoluteness and len(winners) > 1:
        return violations
    
    # For individual voter case or uniform coalition
    if uniform_coalition:
        # Check each ranking type in the profile
        for r in prof.ranking_types:
            # Skip if there aren't enough voters with this ranking for the coalition
            if isinstance(r, tuple):
                if prof.rankings.count(r) < coalition_size:
                    continue
                ranked_candidates = list(r)
                r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
            else:
                if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < coalition_size:
                    continue
                ranked_candidates = list(r.cands)
                r_as_ranking = r
            
            # Try truncating at different positions (keep only first i candidates)
            for i in range(1, len(ranked_candidates)):
                # Create truncated ballot keeping only the first i candidates
                # This is proper truncation - keeping the top i candidates in their original order
                truncated_r = truncate_ballot_from_bottom(r_as_ranking, i)
                
                # Skip if truncation didn't change anything or if truncated ballot is empty
                # Ensure at least one candidate remains ranked
                if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                    continue
                
                # Get the truncated ballot as a ranking map
                truncated_rmap = truncated_r.rmap
                
                # Create a new profile with the truncated ballot(s)
                modified_rankings = []
                
                # Add all ballots except those we'll truncate
                for ballot in prof.rankings:
                    # Skip the ballots we'll truncate
                    if isinstance(ballot, tuple) and ballot == r:
                        continue
                    elif isinstance(ballot, Ranking) and isinstance(r, Ranking) and ballot.cands == r.cands:
                        continue
                    
                    # Add the ballot in the correct format
                    if isinstance(ballot, tuple):
                        modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                    else:
                        modified_rankings.append(ballot.rmap)
                
                # Make sure we've skipped the right number of ballots
                if isinstance(r, tuple):
                    original_count = prof.rankings.count(r)
                else:
                    original_count = sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands)
                
                # Add back any ballots we shouldn't have truncated
                for _ in range(original_count - coalition_size):
                    modified_rankings.append(r_as_ranking.rmap)
                
                # Add the truncated ballots
                for _ in range(coalition_size):
                    modified_rankings.append(truncated_rmap)
                
                # Create the new profile with ties
                new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                    new_prof.use_extended_strict_preference()
                
                # Get the new winners
                new_winners = vm(new_prof)
                
                # Convert numpy array winners to list if needed
                if isinstance(new_winners, np.ndarray):
                    new_winners = new_winners.tolist()
                
                # If require_resoluteness is True, skip profiles with multiple winners after truncation
                if require_resoluteness and len(new_winners) > 1:
                    continue
                
                # Check for Later No Harm violation: a candidate ranked in the truncated ballot
                # goes from losing to winning
                truncated_candidates = truncated_r.cands
                
                # Find candidates that are ranked in the truncated ballot and went from losing to winning
                new_winners_in_truncated = [c for c in new_winners if c in truncated_candidates and c not in winners]
                
                if new_winners_in_truncated:
                    violations.append((r, truncated_r, winners, new_winners, new_winners_in_truncated))
                    if verbose:
                        print(f"Later No Harm violation: Candidate(s) {new_winners_in_truncated} went from losing to winning due to bottom truncation.")
                        print("")
                        print(f"Original winners: {winners}")
                        print(f"New winners: {new_winners}")
                        print(f"Original ballot: {r}")
                        print(f"Truncated ballot: {truncated_r}")
                        print("")
                        print("Original profile:")
                        prof.display()
                        prof.display_margin_graph()
                        vm.display(prof)
                        print("")
                        print("Modified profile:")
                        new_prof.display()
                        new_prof.display_margin_graph()
                        vm.display(new_prof)
    
    # For non-uniform coalition
    if not uniform_coalition and coalition_size > 1:
        # Get all possible combinations of ranking types
        ranking_combinations = list(combinations(prof.ranking_types, coalition_size))
        
        for ranking_combo in ranking_combinations:
            # Check if we have enough of each ranking type
            valid_combo = True
            for r in ranking_combo:
                if isinstance(r, tuple):
                    if prof.rankings.count(r) < ranking_combo.count(r):
                        valid_combo = False
                        break
                else:
                    if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < ranking_combo.count(r):
                        valid_combo = False
                        break
            
            if not valid_combo:
                continue
            
            # For each ranking in the combination, try truncating it
            for i, r in enumerate(ranking_combo):
                if isinstance(r, tuple):
                    ranked_candidates = list(r)
                    r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
                else:
                    ranked_candidates = list(r.cands)
                    r_as_ranking = r
                
                # Try truncating at different positions
                for j in range(1, len(ranked_candidates)):
                    # Create truncated ballot keeping only the first j candidates
                    # This is proper truncation - keeping the top j candidates in their original order
                    truncated_r = truncate_ballot_from_bottom(r_as_ranking, j)
                    
                    # Skip if truncation didn't change anything or if truncated ballot is empty
                    if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                        continue
                    
                    # Get the truncated ballot as a ranking map
                    truncated_rmap = truncated_r.rmap
                    
                    # Create a new profile with the truncated ballot(s)
                    modified_rankings = []
                    
                    # Add all ballots except those in the coalition
                    for ballot in prof.rankings:
                        skip = False
                        for coalition_r in ranking_combo:
                            if (isinstance(ballot, tuple) and ballot == coalition_r) or (isinstance(ballot, Ranking) and isinstance(coalition_r, Ranking) and ballot.cands == coalition_r.cands):
                                skip = True
                                break
                        
                        if skip:
                            continue
                        
                        # Add the ballot in the correct format
                        if isinstance(ballot, tuple):
                            modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                        else:
                            modified_rankings.append(ballot.rmap)
                    
                    # Add back the coalition ballots, with the i-th one truncated
                    for k, coalition_r in enumerate(ranking_combo):
                        if k == i:
                            modified_rankings.append(truncated_rmap)
                        else:
                            if isinstance(coalition_r, tuple):
                                modified_rankings.append({c: i+1 for i, c in enumerate(coalition_r)})
                            else:
                                modified_rankings.append(coalition_r.rmap)
                    
                    # Create the new profile with ties
                    new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                    if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                        new_prof.use_extended_strict_preference()
                    
                    # Get the new winners
                    new_winners = vm(new_prof)
                    
                    # Convert numpy array winners to list if needed
                    if isinstance(new_winners, np.ndarray):
                        new_winners = new_winners.tolist()
                    
                    # If require_resoluteness is True, skip profiles with multiple winners after truncation
                    if require_resoluteness and len(new_winners) > 1:
                        continue
                    
                    # Check for Later No Harm violation: a candidate ranked in the truncated ballot
                    # goes from losing to winning
                    truncated_candidates = truncated_r.cands
                    
                    # Find candidates that are ranked in the truncated ballot and went from losing to winning
                    new_winners_in_truncated = [c for c in new_winners if c in truncated_candidates and c not in winners]
                    
                    if new_winners_in_truncated:
                        violations.append((r, truncated_r, winners, new_winners, new_winners_in_truncated))
                        if verbose:
                            print(f"Later No Harm violation: Candidate(s) {new_winners_in_truncated} went from losing to winning due to bottom truncation.")
                            print("")
                            print(f"Original winners: {winners}")
                            print(f"New winners: {new_winners}")
                            print(f"Original ballot: {r}")
                            print(f"Truncated ballot: {truncated_r}")
                            print("")
                            print("Original profile:")
                            prof.display()
                            prof.display_margin_graph()
                            vm.display(prof)
                            print("")
                            print("Modified profile:")
                            new_prof.display()
                            new_prof.display_margin_graph()
                            vm.display(new_prof)
    
    return violations

later_no_harm = Axiom(
    "Later No Harm",
    has_violation=has_later_no_harm_violation,
    find_all_violations=find_all_later_no_harm_violations
)

def truncate_ballot_from_top(ranking, num_to_keep):
    """
    Truncate a ballot by removing the top candidates and keeping only the bottom num_to_keep candidates.
    
    Args:
        ranking: A Ranking object or tuple representing a ballot
        num_to_keep: Number of bottom candidates to keep
        
    Returns:
        A truncated Ranking object
    """
    if isinstance(ranking, tuple):
        # For tuple rankings, keep only the last num_to_keep candidates
        truncated = ranking[-num_to_keep:]
        return Ranking({c: i+1 for i, c in enumerate(truncated)})
    else:
        # For Ranking objects, sort candidates by rank and keep only the bottom num_to_keep
        sorted_candidates = sorted(ranking.rmap.keys(), key=lambda c: ranking.rmap[c])
        truncated_rmap = {c: ranking.rmap[c] for c in sorted_candidates[-num_to_keep:]}
        return Ranking(truncated_rmap)

def has_earlier_no_help_violation(prof, vm, verbose=False, coalition_size=1, uniform_coalition=True, require_resoluteness=False):
    """
    Returns True if there is a ballot (or collection of ballots) such that by truncating it from the top, 
    a candidate who is ranked by the truncated ballot goes from winning to losing.

    Viewed in reverse, this means that adding previously unranked candidates to the top of a ballot helped lower ranked candidate to win. 

    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.
        coalition_size (int, default=1): Size of the coalition of voters who truncate their ballots.
        uniform_coalition (bool, default=True): If True, all voters in the coalition have the same ballot.
        require_resoluteness (bool, default=False): If True, only profiles with a unique winner before and after truncation are considered.
        
    Returns:
        Result of the test (bool): Returns True if there is a violation and False otherwise.
    """
    # Get the current winners
    winners = vm(prof)
    
    # Convert numpy array winners to list if needed
    if isinstance(winners, np.ndarray):
        winners = winners.tolist()
    
    # If require_resoluteness is True, skip profiles with multiple winners before truncation
    if require_resoluteness and len(winners) > 1:
        return False
    
    # For individual voter case or uniform coalition
    if uniform_coalition:
        # Check each ranking type in the profile
        for r in prof.ranking_types:
            # Skip if there aren't enough voters with this ranking for the coalition
            if isinstance(r, tuple):
                if prof.rankings.count(r) < coalition_size:
                    continue
                ranked_candidates = list(r)
                r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
            else:
                if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < coalition_size:
                    continue
                ranked_candidates = list(r.cands)
                r_as_ranking = r
            
            # Try truncating at different positions (keep only last i candidates)
            for i in range(1, len(ranked_candidates)):
                # Create truncated ballot keeping only the last i candidates
                # This is proper truncation - keeping the bottom i candidates in their original order
                truncated_r = truncate_ballot_from_top(r_as_ranking, i)
                
                # Skip if truncation didn't change anything or if truncated ballot is empty
                # Ensure at least one candidate remains ranked
                if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                    continue
                
                # Get the truncated ballot as a ranking map
                truncated_rmap = truncated_r.rmap
                
                # Create a new profile with the truncated ballot(s)
                modified_rankings = []
                
                # Add all ballots except those we'll truncate
                for ballot in prof.rankings:
                    # Skip the ballots we'll truncate
                    if isinstance(ballot, tuple) and ballot == r:
                        continue
                    elif isinstance(ballot, Ranking) and isinstance(r, Ranking) and ballot.cands == r.cands:
                        continue
                    
                    # Add the ballot in the correct format
                    if isinstance(ballot, tuple):
                        modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                    else:
                        modified_rankings.append(ballot.rmap)
                
                # Make sure we've skipped the right number of ballots
                if isinstance(r, tuple):
                    original_count = prof.rankings.count(r)
                else:
                    original_count = sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands)
                
                # Add back any ballots we shouldn't have truncated
                for _ in range(original_count - coalition_size):
                    modified_rankings.append(r_as_ranking.rmap)
                
                # Add the truncated ballots
                for _ in range(coalition_size):
                    modified_rankings.append(truncated_rmap)
                
                # Create the new profile with ties
                new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                    new_prof.use_extended_strict_preference()
                
                # Get the new winners
                new_winners = vm(new_prof)
                
                # Convert numpy array winners to list if needed
                if isinstance(new_winners, np.ndarray):
                    new_winners = new_winners.tolist()
                
                # If require_resoluteness is True, skip profiles with multiple winners after truncation
                if require_resoluteness and len(new_winners) > 1:
                    continue
                
                # Check for Earlier No Help violation: a candidate ranked in the truncated ballot
                # goes from winning to losing
                truncated_candidates = truncated_r.cands
                
                # Find candidates that are ranked in the truncated ballot and went from winning to losing
                winners_in_truncated = [c for c in winners if c in truncated_candidates]
                new_losers_in_truncated = [c for c in winners_in_truncated if c not in new_winners]
                
                if new_losers_in_truncated:
                    if verbose:
                        print(f"Earlier No Help violation: Candidate(s) {new_losers_in_truncated} went from winning to losing due to top truncation.")
                        print("")
                        print(f"Original winners: {winners}")
                        print(f"New winners: {new_winners}")
                        print(f"Original ballot: {r}")
                        print(f"Truncated ballot: {truncated_r}")
                        print("")
                        print("Original profile:")
                        prof.display()
                        prof.display_margin_graph()
                        vm.display(prof)
                        print("")
                        print("Modified profile:")
                        new_prof.display()
                        new_prof.display_margin_graph()
                        vm.display(new_prof)
                    return True
    
    # For non-uniform coalition
    if not uniform_coalition and coalition_size > 1:
        # Get all possible combinations of ranking types
        ranking_combinations = list(combinations(prof.ranking_types, coalition_size))
        
        for ranking_combo in ranking_combinations:
            # Check if we have enough of each ranking type
            valid_combo = True
            for r in ranking_combo:
                if isinstance(r, tuple):
                    if prof.rankings.count(r) < ranking_combo.count(r):
                        valid_combo = False
                        break
                else:
                    if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < ranking_combo.count(r):
                        valid_combo = False
                        break
            
            if not valid_combo:
                continue
            
            # For each ranking in the combination, try truncating it
            for i, r in enumerate(ranking_combo):
                if isinstance(r, tuple):
                    ranked_candidates = list(r)
                    r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
                else:
                    ranked_candidates = list(r.cands)
                    r_as_ranking = r
                
                # Try truncating at different positions
                for j in range(1, len(ranked_candidates)):
                    # Create truncated ballot keeping only the last j candidates
                    # This is proper truncation - keeping the bottom j candidates in their original order
                    truncated_r = truncate_ballot_from_top(r_as_ranking, j)
                    
                    # Skip if truncation didn't change anything or if truncated ballot is empty
                    if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                        continue
                    
                    # Get the truncated ballot as a ranking map
                    truncated_rmap = truncated_r.rmap
                    
                    # Create a new profile with the truncated ballot(s)
                    modified_rankings = []
                    
                    # Add all ballots except those in the coalition
                    for ballot in prof.rankings:
                        skip = False
                        for coalition_r in ranking_combo:
                            if (isinstance(ballot, tuple) and ballot == coalition_r) or (isinstance(ballot, Ranking) and isinstance(coalition_r, Ranking) and ballot.cands == coalition_r.cands):
                                skip = True
                                break
                        
                        if skip:
                            continue
                        
                        # Add the ballot in the correct format
                        if isinstance(ballot, tuple):
                            modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                        else:
                            modified_rankings.append(ballot.rmap)
                    
                    # Add back the coalition ballots, with the i-th one truncated
                    for k, coalition_r in enumerate(ranking_combo):
                        if k == i:
                            modified_rankings.append(truncated_rmap)
                        else:
                            if isinstance(coalition_r, tuple):
                                modified_rankings.append({c: i+1 for i, c in enumerate(coalition_r)})
                            else:
                                modified_rankings.append(coalition_r.rmap)
                    
                    # Create the new profile with ties
                    new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                    if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                        new_prof.use_extended_strict_preference()
                    
                    # Get the new winners
                    new_winners = vm(new_prof)
                    
                    # Convert numpy array winners to list if needed
                    if isinstance(new_winners, np.ndarray):
                        new_winners = new_winners.tolist()
                    
                    # If require_resoluteness is True, skip profiles with multiple winners after truncation
                    if require_resoluteness and len(new_winners) > 1:
                        continue
                    
                    # Check for Earlier No Help violation: a candidate ranked in the truncated ballot
                    # goes from winning to losing
                    truncated_candidates = truncated_r.cands
                    
                    # Find candidates that are ranked in the truncated ballot and went from winning to losing
                    winners_in_truncated = [c for c in winners if c in truncated_candidates]
                    new_losers_in_truncated = [c for c in winners_in_truncated if c not in new_winners]
                    
                    if new_losers_in_truncated:
                        if verbose:
                            print(f"Earlier No Help violation: Candidate(s) {new_losers_in_truncated} went from winning to losing due to top truncation.")
                            print("")
                            print(f"Original winners: {winners}")
                            print(f"New winners: {new_winners}")
                            print(f"Original ballot: {r}")
                            print(f"Truncated ballot: {truncated_r}")
                            print("")
                            print("Original profile:")
                            prof.display()
                            prof.display_margin_graph()
                            vm.display(prof)
                            print("")
                            print("Modified profile:")
                            new_prof.display()
                            new_prof.display_margin_graph()
                            vm.display(new_prof)
                        return True
    
    return False

def find_all_earlier_no_help_violations(prof, vm, verbose=False, coalition_size=1, uniform_coalition=True, require_resoluteness=False):
    """
    Returns a list of tuples (original_ballot, truncated_ballot, original_winners, new_winners, new_losers_in_truncated) such that top-truncating the original_ballot to the truncated_ballot causes a candidate ranked in the truncated ballot to go from winning to losing.
    
    Args:
        prof: a Profile or ProfileWithTies object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation.
        coalition_size (int, default=1): Size of the coalition of voters who truncate their ballots.
        uniform_coalition (bool, default=True): If True, all voters in the coalition have the same ballot.
        require_resoluteness (bool, default=False): If True, only profiles with a unique winner before and after truncation are considered.
        
    Returns:
        A list of tuples (original_ballot, truncated_ballot, original_winners, new_winners, new_losers_in_truncated) 
        witnessing violations of Earlier No Help. The new_losers_in_truncated element contains the candidates that 
        specifically caused the violation by going from winning to losing while being ranked in the truncated ballot.
    """
    violations = []
    
    # Get the current winners
    winners = vm(prof)
    
    # Convert numpy array winners to list if needed
    if isinstance(winners, np.ndarray):
        winners = winners.tolist()
    
    # If require_resoluteness is True, skip profiles with multiple winners before truncation
    if require_resoluteness and len(winners) > 1:
        return violations
    
    # For individual voter case or uniform coalition
    if uniform_coalition:
        # Check each ranking type in the profile
        for r in prof.ranking_types:
            # Skip if there aren't enough voters with this ranking for the coalition
            if isinstance(r, tuple):
                if prof.rankings.count(r) < coalition_size:
                    continue
                ranked_candidates = list(r)
                r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
            else:
                if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < coalition_size:
                    continue
                ranked_candidates = list(r.cands)
                r_as_ranking = r
            
            # Try truncating at different positions (keep only last i candidates)
            for i in range(1, len(ranked_candidates)):
                # Create truncated ballot keeping only the last i candidates
                # This is proper truncation - keeping the bottom i candidates in their original order
                truncated_r = truncate_ballot_from_top(r_as_ranking, i)
                
                # Skip if truncation didn't change anything or if truncated ballot is empty
                # Ensure at least one candidate remains ranked
                if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                    continue
                
                # Get the truncated ballot as a ranking map
                truncated_rmap = truncated_r.rmap
                
                # Create a new profile with the truncated ballot(s)
                modified_rankings = []
                
                # Add all ballots except those we'll truncate
                for ballot in prof.rankings:
                    # Skip the ballots we'll truncate
                    if isinstance(ballot, tuple) and ballot == r:
                        continue
                    elif isinstance(ballot, Ranking) and isinstance(r, Ranking) and ballot.cands == r.cands:
                        continue
                    
                    # Add the ballot in the correct format
                    if isinstance(ballot, tuple):
                        modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                    else:
                        modified_rankings.append(ballot.rmap)
                
                # Make sure we've skipped the right number of ballots
                if isinstance(r, tuple):
                    original_count = prof.rankings.count(r)
                else:
                    original_count = sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands)
                
                # Add back any ballots we shouldn't have truncated
                for _ in range(original_count - coalition_size):
                    modified_rankings.append(r_as_ranking.rmap)
                
                # Add the truncated ballots
                for _ in range(coalition_size):
                    modified_rankings.append(truncated_rmap)
                
                # Create the new profile with ties
                new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                    new_prof.use_extended_strict_preference()
                
                # Get the new winners
                new_winners = vm(new_prof)
                
                # Convert numpy array winners to list if needed
                if isinstance(new_winners, np.ndarray):
                    new_winners = new_winners.tolist()
                
                # If require_resoluteness is True, skip profiles with multiple winners after truncation
                if require_resoluteness and len(new_winners) > 1:
                    continue
                
                # Check for Earlier No Help violation: a candidate ranked in the truncated ballot
                # goes from winning to losing
                truncated_candidates = truncated_r.cands
                
                # Find candidates that are ranked in the truncated ballot and went from winning to losing
                winners_in_truncated = [c for c in winners if c in truncated_candidates]
                new_losers_in_truncated = [c for c in winners_in_truncated if c not in new_winners]
                
                if new_losers_in_truncated:
                    violations.append((r, truncated_r, winners, new_winners, new_losers_in_truncated))
                    if verbose:
                        print(f"Earlier No Help violation: Candidate(s) {new_losers_in_truncated} went from winning to losing due to top truncation.")
                        print("")
                        print(f"Original winners: {winners}")
                        print(f"New winners: {new_winners}")
                        print(f"Original ballot: {r}")
                        print(f"Truncated ballot: {truncated_r}")
                        print("")
                        print("Original profile:")
                        prof.display()
                        prof.display_margin_graph()
                        vm.display(prof)
                        print("")
                        print("Modified profile:")
                        new_prof.display()
                        new_prof.display_margin_graph()
                        vm.display(new_prof)
    
    # For non-uniform coalition
    if not uniform_coalition and coalition_size > 1:
        # Get all possible combinations of ranking types
        ranking_combinations = list(combinations(prof.ranking_types, coalition_size))
        
        for ranking_combo in ranking_combinations:
            # Check if we have enough of each ranking type
            valid_combo = True
            for r in ranking_combo:
                if isinstance(r, tuple):
                    if prof.rankings.count(r) < ranking_combo.count(r):
                        valid_combo = False
                        break
                else:
                    if sum(1 for ballot in prof.rankings if isinstance(ballot, Ranking) and ballot.cands == r.cands) < ranking_combo.count(r):
                        valid_combo = False
                        break
            
            if not valid_combo:
                continue
            
            # For each ranking in the combination, try truncating it
            for i, r in enumerate(ranking_combo):
                if isinstance(r, tuple):
                    ranked_candidates = list(r)
                    r_as_ranking = Ranking({c: i+1 for i, c in enumerate(r)})
                else:
                    ranked_candidates = list(r.cands)
                    r_as_ranking = r
                
                # Try truncating at different positions
                for j in range(1, len(ranked_candidates)):
                    # Create truncated ballot keeping only the last j candidates
                    # This is proper truncation - keeping the bottom j candidates in their original order
                    truncated_r = truncate_ballot_from_top(r_as_ranking, j)
                    
                    # Skip if truncation didn't change anything or if truncated ballot is empty
                    if len(truncated_r.cands) == 0 or truncated_r.rmap == r_as_ranking.rmap:
                        continue
                    
                    # Get the truncated ballot as a ranking map
                    truncated_rmap = truncated_r.rmap
                    
                    # Create a new profile with the truncated ballot(s)
                    modified_rankings = []
                    
                    # Add all ballots except those in the coalition
                    for ballot in prof.rankings:
                        skip = False
                        for coalition_r in ranking_combo:
                            if (isinstance(ballot, tuple) and ballot == coalition_r) or (isinstance(ballot, Ranking) and isinstance(coalition_r, Ranking) and ballot.cands == coalition_r.cands):
                                skip = True
                                break
                        
                        if skip:
                            continue
                        
                        # Add the ballot in the correct format
                        if isinstance(ballot, tuple):
                            modified_rankings.append({c: i+1 for i, c in enumerate(ballot)})
                        else:
                            modified_rankings.append(ballot.rmap)
                    
                    # Add back the coalition ballots, with the i-th one truncated
                    for k, coalition_r in enumerate(ranking_combo):
                        if k == i:
                            modified_rankings.append(truncated_rmap)
                        else:
                            if isinstance(coalition_r, tuple):
                                modified_rankings.append({c: i+1 for i, c in enumerate(coalition_r)})
                            else:
                                modified_rankings.append(coalition_r.rmap)
                    
                    # Create the new profile with ties
                    new_prof = ProfileWithTies(modified_rankings, candidates=prof.candidates)
                    if isinstance(prof, ProfileWithTies) and prof.using_extended_strict_preference:
                        new_prof.use_extended_strict_preference()
                    
                    # Get the new winners
                    new_winners = vm(new_prof)
                    
                    # Convert numpy array winners to list if needed
                    if isinstance(new_winners, np.ndarray):
                        new_winners = new_winners.tolist()
                    
                    # If require_resoluteness is True, skip profiles with multiple winners after truncation
                    if require_resoluteness and len(new_winners) > 1:
                        continue
                    
                    # Check for Earlier No Help violation: a candidate ranked in the truncated ballot
                    # goes from winning to losing
                    truncated_candidates = truncated_r.cands
                    
                    # Find candidates that are ranked in the truncated ballot and went from winning to losing
                    winners_in_truncated = [c for c in winners if c in truncated_candidates]
                    new_losers_in_truncated = [c for c in winners_in_truncated if c not in new_winners]
                    
                    if new_losers_in_truncated:
                        violations.append((r, truncated_r, winners, new_winners, new_losers_in_truncated))
                        if verbose:
                            print(f"Earlier No Help violation: Candidate(s) {new_losers_in_truncated} went from winning to losing due to top truncation.")
                            print("")
                            print(f"Original winners: {winners}")
                            print(f"New winners: {new_winners}")
                            print(f"Original ballot: {r}")
                            print(f"Truncated ballot: {truncated_r}")
                            print("")
                            print("Original profile:")
                            prof.display()
                            prof.display_margin_graph()
                            vm.display(prof)
                            print("")
                            print("Modified profile:")
                            new_prof.display()
                            new_prof.display_margin_graph()
                            vm.display(new_prof)
    
    return violations

earlier_no_help = Axiom(
    "Earlier No Help",
    has_violation=has_earlier_no_help_violation,
    find_all_violations=find_all_earlier_no_help_violations
)


"""
    File: dominance_axioms.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: April 27, 2023
    
    Dominance axioms 
"""

from pref_voting.axiom import Axiom
from pref_voting.c1_methods import smith_set, schwartz_set
from pref_voting.axiom_helpers import *


def has_pareto_dominance_violation(edata, vm, verbose=False, strong_Pareto = False):
    """
    Returns True if some winner according to vm is Pareto dominated (there is a candidate that is unanimously preferred to the winner).

    If strong_Pareto is True, then a candidate A is dominated if there is a candidate B such that some voter prefers B to A and no voter prefers A to B.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    ws = vm(edata)
    for w in ws: 
        for c in edata.candidates: 
            if (strong_Pareto == False and edata.support(c,w)==edata.num_voters) or (strong_Pareto == True and edata.support(c,w)> 0 and edata.support(w,c)==0):
                if verbose:  
                    print(f"Pareto violation by {vm}:")
                    edata.display()
                    print(edata.description())
                    vm.display(edata)
                    if not strong_Pareto:
                        print(f"Every voter ranks {c} higher than the winner {w}.")
                    else:
                        print(f"Every voter ranks {c} at least as highly as the winner {w}, and some voter ranks {c} strictly higher than {w}.")
                    print()
                return True
    return False

def find_all_pareto_dominance_violations(edata, vm, verbose=False, strong_Pareto = False):
    """
    Returns all Pareto-dominated winners.

    If strong_Pareto is True, then a candidate A is dominated if there is a candidate B such that some voter prefers B to A and no voter prefers A to B.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        pareto_dominated_winners: A list of tuples of candidates (w, c) where w is a winner according to vm and Pareto dominated by c (and the empty list if there are none). 

    """

    ws = vm(edata)
    pareto_dominated_winners = list()
    for w in ws: 
        for c in edata.candidates: 
            if (strong_Pareto == False and edata.support(c,w)==edata.num_voters) or (strong_Pareto == True and edata.support(c,w)> 0 and edata.support(w,c)==0):
                pareto_dominated_winners.append((w, c))
    
    if len(pareto_dominated_winners) > 0 and verbose: 
        print(f"Pareto violation by {vm}:")
        edata.display()
        print(edata.description())
        vm.display(edata)
        for w,c in pareto_dominated_winners:
            if not strong_Pareto:
                print(f"Every voter ranks {c} higher than the winner {w}.")
            else:
                print(f"Every voter ranks {c} at least as highly as the winner {w}, and some voter ranks {c} strictly higher than {w}.")

    return pareto_dominated_winners

pareto_dominance = Axiom(
    "Pareto Dominance Criterion",
    has_violation = has_pareto_dominance_violation,
    find_all_violations = find_all_pareto_dominance_violations, 
)

def has_condorcet_winner_violation(edata, vm, only_resolute=False, verbose=False):
    """
    Returns True if there is a Condorcet winner in edata (a candidate that is majority preferred to every other candidate) that is not the unique winner according to vm.
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        only_resolute (bool, default=False): If True, only consider profiles in which there is a unique winner according to vm
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    cw = edata.condorcet_winner()

    ws = vm(edata)
    
    if only_resolute and len(ws) != 1: 
        return False
    if cw is not None and ws != [cw]:
        if verbose:
            print("Violation of Condorcet Winner by", vm.name,"on profile:")
            if isinstance(edata, (Profile, ProfileWithTies)):
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else:
                edata.display()
                print(edata.description())
            
            print(f"The Condorcet winner {cw} is not the unique winner: ")
            vm.display(edata)
        return True
    return False

def find_all_condorcet_winner_violations(edata, vm, only_resolute=False, verbose=False):
    """
    Returns the Condorcet winner that is not the unique winner according to vm.
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        only_resolute (bool, default=False): If True, only consider profiles in which there is a unique winner according to vm.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    cw = edata.condorcet_winner()

    ws = vm(edata)

    if only_resolute and len(ws) != 1: 
        return list()

    if cw is not None and ws != [cw]:
        if verbose: 
            print("Violation of Condorcet Winner by", vm.name,"on profile:")
            if isinstance(edata, (Profile, ProfileWithTies)):
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else:
                edata.display()
                print(edata.description())

            print(f"The Condorcet winner {cw} is not the unique winner: ")
            vm.display(edata)
        return [cw] 
    return list()

condorcet_winner = Axiom(
    "Condorcet Winner",
    has_violation = has_condorcet_winner_violation,
    find_all_violations = find_all_condorcet_winner_violations, 
)

def has_condorcet_loser_violation(edata, vm, verbose=False):
    """
    Returns True if there is a winner according to vm that is a Condorcet loser (a candidate that loses head-to-head to every other candidate).  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    cl = edata.condorcet_loser()

    ws = vm(edata)

    if cl is not None and cl in ws:
        if verbose:
            print("Violation of Condorcet Loser by", vm.name,"on profile:")
            if type(edata) == Profile or type(edata) == ProfileWithTies:
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else: 
                edata.display()
                print(edata.description())

            print(f"The Condorcet loser {cl} is an element of the winning set: ")
            vm.display(edata)
        return True 
    return False

def find_all_condorcet_loser_violations(edata, vm, verbose=False):
    """
    Returns the Condorcet loser (a candidate that loses head-to-head to every other candidate) who is a winner according to vm.  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    cl = edata.condorcet_loser()

    ws = vm(edata)

    if cl is not None and cl in ws:
        if verbose: 
            print("Violation of Condorcet Loser by", vm.name,"on profile:")
            if type(edata) == Profile or type(edata) == ProfileWithTies: 
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else: 
                edata.display()
                print(edata.description())
 
            print(f"The Condorcet loser {cl} is an element of the winning set: ")
            vm.display(edata)
        return [cl] 
    return list()

condorcet_loser = Axiom(
    "Condorcet Loser",
    has_violation = has_condorcet_loser_violation,
    find_all_violations = find_all_condorcet_loser_violations, 
)

def has_smith_violation(edata, vm, verbose=False):
    """
    Returns True if there is a winner according to vm that is not in the Smith set (the smallest set of candidates such that every candidate in the set is majority preferred to every candidate outside the set).  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    s_set = smith_set(edata)
    ws = vm(edata)

    winners_not_in_smith = [w for w in ws if w not in s_set]
    if len(winners_not_in_smith) > 0: 
        if verbose:
            print("Violation of Smith by", vm.name,"on profile:")
            if type(edata) == Profile or type(edata) == ProfileWithTies:
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else: 
                edata.display()
                print(edata.description())
            print("Smith set:", list_to_string(s_set, edata.cmap))
            print(f"The winners who are not in the Smith set: {list_to_string(winners_not_in_smith, edata.cmap)}")
            vm.display(edata)
        return True 
    return False

def find_all_smith_violations(edata, vm, verbose=False):
    """
    Returns the winners according to vm that are not in the Smith set (the smallest set of candidates such that every candidate in the set is majority preferred to every candidate outside the set).  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    s_set = smith_set(edata)
    ws = vm(edata)

    winners_not_in_smith = [w for w in ws if w not in s_set]
    if len(winners_not_in_smith) > 0: 
        if verbose:
            if type(edata) == Profile or type(edata) == ProfileWithTies: 
                edata.display()
                print(edata.description())
                edata.display_margin_graph()
            else: 
                edata.display()
                print(edata.description())
            print("Smith set:", list_to_string(s_set, edata.cmap))
            print(f"The winners who are not in the Smith set: {list_to_string(winners_not_in_smith, edata.cmap)}")
            vm.display(edata)
        return winners_not_in_smith 
    return list()

smith = Axiom(
    "Smith",
    has_violation = has_smith_violation,
    find_all_violations = find_all_smith_violations, 
)

def has_schwartz_violation(edata, vm, verbose=False):
    """
    Returns True if there is a winner according to vm that is not in the Schwartz set (the set of all candidates x such that if y can reach x in the transitive closer of the majority relation, then x can reach y in the transitive closer of the majority relation.).  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    s_set = schwartz_set(edata)
    ws = vm(edata)

    winners_not_in_schwartz = [w not in s_set for w in ws]
    if len(winners_not_in_schwartz) > 0: 
        if verbose:
            if type(edata) == Profile or type(edata) == ProfileWithTies: 
                edata.display_margin_graph()
            else: 
                edata.display()
            print(f"The winners that are not in the Schwartz set: {list_to_string(winners_not_in_schwartz, edata.cmap)}.")
            vm.display(edata)
        return True 
    return False

def find_all_schwartz_violations(edata, vm, verbose=False):
    """
    Returns the winners according to vm that are not in the Schwartz set (the set of all candidates x such that if y can reach x in the transitive closer of the majority relation, then x can reach y in the transitive closer of the majority relation).  
    
    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, or MarginGraph): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """

    s_set = schwartz_set(edata)
    ws = vm(edata)

    winners_not_in_schwartz = [w not in s_set for w in ws]
    if len(winners_not_in_schwartz) > 0: 
        if verbose:
            if type(edata) == Profile or type(edata) == ProfileWithTies: 
                edata.display_margin_graph()
            else: 
                edata.display()
            print(f"The winners that are not in the Schwartz set: {list_to_string(winners_not_in_schwartz, edata.cmap)}.")
            vm.display(edata)
        return winners_not_in_schwartz 
    return list()

schwartz = Axiom(
    "Schwartz",
    has_violation = has_schwartz_violation,
    find_all_violations = find_all_schwartz_violations, 
)

dominance_axioms = [
    pareto_dominance, 
    condorcet_winner, 
    condorcet_loser,
    smith,
    schwartz
]


"""
    File: readers.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: March 17, 2024

    Functions to write election data to a file.
"""

from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.spatial_profiles import SpatialProfile
from preflibtools.instances import OrdinalInstance
import os
import csv
import pandas as pd
import json

def abif_to_profile(filename):
    """
    Open filename in the abif format and return a Profile object.

    Args:
        filename: The name of the file to read the profile from.
    
    Returns:
        A Profile object.

    """

    with open(filename, mode='r') as file:
        
        lines = list(file.readlines())

        cmap = {}
        cand_to_indices = {}
        cindx = 0
        # create a candidate map
        for line in lines: 
            if line.startswith("="):
                _, cname = line[1:].strip().split(":")
                cmap[cindx] = cname.strip().strip("[]")
                cand_to_indices[cname.strip().strip("[]")] = cindx
                cindx += 1

        rankings = []
        rcounts = []
        for line in lines:
            if line.startswith("#"):
                # comment
                continue
            elif line.startswith("="):
                # candidate line
                continue
            elif line.startswith("{"):
                # metadata
                continue
            else:
                # ranking line
                count, ranking = line.strip().split(":")
                count = int(count)
                ranking = ranking.split(">")

                assert not any(["=" in cs or "," in cs for cs in ranking]), "The election must contain linear orders on the candidates to create a Profile."
                
                if len(cmap) == 0:
                    # no candidate map provided, so need to create one from the rankings
                    cmap = {cidx: str(sorted(ranking)[cidx].strip()) for cidx in range(len(sorted(ranking)))}
                    cand_to_indices = {c: i for i, c in cmap.items()}
                
                r = list()
                assert len(cmap) > 0 and len(ranking) == len(cmap), "The election must contain linear orders on the candidates to create a Profile."
                for c in ranking:
                    assert len(cmap) > 0 and c in cand_to_indices.keys(), "Candidate found that is not in the candidate map."
                    r.append(cand_to_indices[c.strip()])
                rankings.append(r)
                rcounts.append(count)

        return Profile(
            rankings, 
            rcounts=rcounts, 
            cmap=cmap)           


def abif_to_profile_with_ties(filename, cand_type=None):
    """
    Open filename in the abif format and return a ProfileWithTies object.

    Args:
        filename: The name of the file to read the profile from.
    
    Returns:
        A ProfileWithTies object.

    """
    
    import re

    with open(filename, mode='r') as file:
        lines = list(file.readlines())
        rankings = []
        rcounts = []
        cmap = {}
        for line in lines:
            if line.startswith("#"):
                # comment
                continue
            elif line.startswith("="):
                # candidate line
                cidx, cname = line[1:].strip().split(":")
                cmap[cand_type(cidx.strip()) 
                    if cand_type is not None else cidx.strip()] = cname.strip().strip("[]")
            elif line.startswith("{"):
                # metadata
                continue
            else:
                # ranking line
                count, ranking = line.strip().split(":")
                count = int(count)
                ranking = ranking.split(">")
                r = dict()
                for ridx, cs in enumerate(ranking):
                    cands = re.split(r'[=,]', cs)
                    for c in cands: 
                        if cand_type is not None: 
                            r[cand_type(c.strip())] = ridx + 1
                        else:
                            r[c.strip()] = ridx + 1
                rankings.append(r)
                rcounts.append(count)

        if len(cmap) == 0:
            return ProfileWithTies(
                rankings, 
                rcounts=rcounts)           
        else:
            return ProfileWithTies(
                rankings, 
                rcounts=rcounts, 
                candidates = sorted(list(cmap.keys())),
                cmap = cmap)           

def preflib_to_profile(
        instance_or_preflib_file, 
        include_cmap=False,
        use_cand_names=False,
        as_linear_profile=False): 
    
    """
    Read a profile from an OrdinalInstance or a .soc, .soi, .toc, or .toi file used by PrefLib (https://www.preflib.org/format#types).

    This function uses the ``OrdinalInstance`` class from the ``preflibtools`` package to read the profile from the file (see https://preflib.github.io/preflibtools/usage.html#ordinal-preferences).

    Args:
        preflib_file (str): the path to the file
        include_cmap (bool): if True, then include the candidate map.  Defaults to False.
        use_cand_names (bool): if True, then use the candidate map as the candidate names.  Defaults to False.
        as_linear_profile (bool): if True, then return a Profile object.  Defaults to False.  If False, then return a ProfileWithTies object.

    Returns:    
        Profile or ProfileWithTies: the profile read from the file
        
    """

    assert type(instance_or_preflib_file) == OrdinalInstance or type(instance_or_preflib_file) == str, "The argument must be an instance of OrdinalInstance or a string."

    if type(instance_or_preflib_file) == str:
        preflib_file = instance_or_preflib_file

        assert preflib_file.endswith(".soc") or preflib_file.endswith(".soi") or preflib_file.endswith(".toc") or preflib_file.endswith(".toi"), f"The file must be one of the file types from preflib: https://www.preflib.org/format#types, not {preflib_file}."

        assert os.path.exists(preflib_file), f"The file {preflib_file} does not exist."

        instance = OrdinalInstance()
        instance.parse_file(preflib_file)

    else:
        instance = instance_or_preflib_file

    rankings = []
    rcounts = []
    cmap = {c:str(c) for c in instance.alternatives_name.keys()}

    if not as_linear_profile:

        for order in instance.orders:
            rank = dict()
            for r,cs in enumerate(order): 
                for c in cs: 
                    if not use_cand_names:
                        rank[c] = r + 1
                    else: 
                        rank[instance.alternatives_name[c]] = r + 1
                    if include_cmap:
                        if  use_cand_names:
                            cmap[instance.alternatives_name[c]] = instance.alternatives_name[c]
                        else:
                            cmap[c] = instance.alternatives_name[c]

            rankings.append(rank)
            rcounts.append(instance.multiplicity[order])

        return ProfileWithTies(rankings, 
                               rcounts=rcounts,
                               cmap=cmap)

    elif as_linear_profile: 
        
        cand_to_cidx = {c:cidx 
                        for cidx,c in enumerate(sorted(list(instance.alternatives_name.keys())))}

        for order in instance.orders:    
            rank = list()
            cmap = {c:str(c) for c in range(instance.num_alternatives)}
            for _,cs in enumerate(order): 
                for c in cs: 
                    rank.append(cand_to_cidx[c])
                    if include_cmap:
                        cmap[cand_to_cidx[c]] = instance.alternatives_name[c]
            rankings.append(rank)
            rcounts.append(instance.multiplicity[order])

        return Profile(rankings, 
                       rcounts=rcounts,
                       cmap=cmap)

def csv_to_profile(
        filename, 
        csv_format="candidate_columns", 
        as_linear_profile=False,
        items_to_skip=None, 
        cand_type=None):
        """
        Read a profile from a csv file. 

        Args:
            filename (str): the path to the file
            csv_format (str): the format of the csv file.  Defaults to "candidate_columns".  The other option is "rank_columns".
            as_linear_profile (bool): if True, then return a Profile object.  Defaults to False.  If False, then return a ProfileWithTies object.
            items_to_skip (list[str]): a list of items to skip.  Defaults to None.  Items in this list are not included in the profile.  Only relevant for "rank_columns" csv format.

        Returns:
            Profile or ProfileWithTies: the profile read from the file

        Note: 
            There are two formats for the csv file: "rank_columns" and "candidate_columns".  The "rank_columns" format is used when the csv file contains a column for each rank and the rows are the candidates at that rank (or "skipped" if the ranked is skipped).  The "candidate_columns" format is used when the csv file contains a column for each candidate and the rows are the rank of the candidates (or the empty string if the candidate is not ranked).
        """
        
        if csv_format == "rank_columns":
            df = pd.read_csv(filename, low_memory=False)
            items_to_skip = items_to_skip if items_to_skip is not None else ["skipped"]
            ranks = []
            rank_columns = [col for col in df.columns if col.startswith('rank') or col.startswith('Rank')]

            # Get unique values from these columns, excluding 'skipped'
            cand_names = pd.unique(df[rank_columns].values.ravel('K'))
            cand_names = [str(value) for value in cand_names if value not in items_to_skip]

            if 'writein' in cand_names:
                cands = list(set([c for c in sorted(cand_names) if c != 'writein'])) + ['writein']
            else: 
                cands = sorted(list(set(cand_names)))
            if len(cands) == 0: 
                print("No candidates found in file", filename)
            cmap = {cidx: c for cidx,c in enumerate(cands)}
            cand_to_cidx = {c:cidx for cidx,c in enumerate(cands)}

            rank_str_to_rank = lambda rank_str: int(rank_str[4:].strip())
            for _, row in df.iterrows():
                ballot_dict = {}
                for rank in rank_columns:
                    candidate = str(row[rank])
                    if candidate not in items_to_skip:
                        ballot_dict[cand_to_cidx[candidate]] = rank_str_to_rank(rank)
                        
                ballot_dict = {cand_type(c) if cand_type is not None else c:r 
                               for c,r in ballot_dict.items()}
                ranks.append(ballot_dict)
            cmap = {cand_to_cidx[c]:str(c) for c in cands}
            prof = ProfileWithTies(ranks, cmap=cmap)
            if as_linear_profile:
                prof = prof.to_linear_profile() 
                assert prof is not None, "The profile could not be converted to a Profile."
            return prof
        
        elif csv_format == "candidate_columns":             
            with open(filename, mode='r') as file:
                reader = csv.reader(file)
                header = next(reader)
                candidates = header[:-1]
                rankings = list()
                rcounts = list()
                for row in reader:
                    ranks = [int(r) if r != "" else None for r in row[:-1]]
                    count = int(row[-1])
                    ranking = {cand_type(c) 
                               if cand_type is not None else c:r 
                               for c,r in zip(candidates, ranks) 
                               if r is not None}
                    rankings.append(ranking)
                    rcounts.append(count)

            prof = ProfileWithTies(rankings, 
                                   rcounts=rcounts, 
                                   cmap={cand_type(c) 
                                         if cand_type is not None else str(c):str(c) 
                                         for c in candidates})
            if as_linear_profile:
                prof = prof.to_linear_profile() 
                assert prof is not None, "The profile could not be converted to a Profile."
            return prof


# helper function for json_to_profile
def _convert_key_type(key, lst):
    for c in lst:
        try:
            # Attempt to convert the key to the same type as the candidate
            if type(c)(key) == c:
                return type(c)(key)
        except ValueError:
            continue
    # Return the original key if no conversion is successful
    return key


def json_to_profile(filename, cand_type=None, as_linear_profile=False): 
    """
    Read a profile from a json file. 

    Args:
        filename (str): the path to the file
        cand_type (type): the type of the candidates.  Defaults to None.  If not None, then the candidates are converted to this type.
        as_linear_profile (bool): if True, then return a Profile object.  Defaults to False.  If False, then return a ProfileWithTies object.

    Returns:
        Profile or ProfileWithTies: the profile read from the file
    """
    with open(filename, mode='r') as file:
        data = json.load(file)
        candidates = data["candidates"]
        cmap = {_convert_key_type(c, candidates): c_str for c, c_str in data["cmap"].items()}

        if cand_type is not None: 
            cmap = {cand_type(c):str(c_str) for c,c_str in cmap.items()}
            candidates = [cand_type(c) for c in candidates]

        rankings = []
        rcounts = []
        for r_data in data["rankings"]:
            rank = {cand_type(c) if cand_type is not None else _convert_key_type(c, candidates):int(r) for c,r in r_data["ranking"].items()}
            rankings.append(rank)
            rcounts.append(int(r_data["count"]))

    if as_linear_profile:
        prof = ProfileWithTies(rankings, 
                               rcounts=rcounts, 
                               candidates=candidates,
                               cmap=cmap)
        
        prof = prof.to_linear_profile() 
        assert prof is not None, "The profile could not be converted to a Profile."
    else:
        prof = ProfileWithTies(rankings, 
                               rcounts=rcounts,
                               candidates=candidates, 
                               cmap=cmap)
    return prof

def read(filename,
         file_format,
         as_linear_profile=False, 
         cand_type=None,
         csv_format="candidate_columns",
         items_to_skip=None): 
    """
    Read election data from ``filename`` in the format ``file_format``. 

    Args:
        filename (str): the path to the file
        file_format (str): the format of the file.  The options are "preflib", "json", "csv", and "abif".
        as_linear_profile (bool): if True, then return a Profile object.  Defaults to False.  If False, then return a ProfileWithTies object.
        cand_type (type): the type of the candidates.  Defaults to None.  If not None, then the candidates are converted to this type.
        csv_format (str): the format of the csv file.  Defaults to "candidate_columns".  The other option is "rank_columns".
        items_to_skip (list[str]): a list of items to skip.  Defaults to None.  Items in this list are not included in the profile.  Only relevant for "rank_columns" csv format.

    Returns:
        Profile or ProfileWithTies: the profile read from the file
    """
    if file_format == "abif":
        if as_linear_profile:
            return abif_to_profile(
                filename)
        else: 
            return abif_to_profile_with_ties(
                filename,
                cand_type=cand_type)
    elif file_format == "json":
        return json_to_profile(
            filename, 
            cand_type=cand_type, 
            as_linear_profile=as_linear_profile)
    elif file_format == "csv":
        return csv_to_profile(
            filename,
            as_linear_profile=as_linear_profile,
            cand_type=cand_type, 
            csv_format=csv_format,
            items_to_skip=items_to_skip)
    elif file_format == "preflib":
        return preflib_to_profile(filename, as_linear_profile=as_linear_profile)
    else:
        raise ValueError(f"File format {file_format} not recognized.")
    
def json_to_spatial_profile(filename): 
    """
    Load a spatial profile from a JSON file.

    Args:
        filename (str): the path to the file

    Returns:
        SpatialProfile: the spatial profile read from the file
    """

    with open(filename, "r") as f:
        spatial_profile_dict = json.load(f)
        candidates = spatial_profile_dict["cand_names"]
        voters = spatial_profile_dict["voter_names"]
        return SpatialProfile(
            {_convert_key_type(c, candidates):c_pos for c,c_pos in spatial_profile_dict["candidates"].items()}, 
            {_convert_key_type(v, voters):v_pos for v,v_pos in spatial_profile_dict["voters"].items()}
            )
